
import math
import random

import pyglet
from pyglet.gl import *
from pyglet.window import key, mouse

from .pytmx import pytmx
from .pytmx.util_pyglet import load_pyglet

from . import menu
from . import hud
from . import player
from . import particle
from . import twin
from . import critter

from .resources import asset_definitions as ad

import attr

@attr.s
class GameState:
    critters_saved: int = attr.ib(default=0)
    critters_killed: int = attr.ib(default=0)

class Game:
    WINDOW_WIDTH = 1600
    WINDOW_HEIGHT = 900
    WINDOW_MIN_SIZE = (640, 480)
    GAME_TITLE = 'My Evil Twin: Save the critters!'

    def __init__(self):
        self.configure_window()

        self.paused = True

        self.audio_player = pyglet.media.Player()
        self.audio_player.volume = 0.5
        self.audio_player.play()  # Start playing anything that gets queued.

        self.music_player = pyglet.media.Player()
        self.music_player.volume = 0.5

        self.map = load_pyglet(ad.MAP_FILE)
        self.map_batch = pyglet.graphics.Batch()
        self.map_sprites = {}
        for layer in self.map.visible_layers:
            for i, (x, y, gid) in enumerate(layer):
                tile = self.map.get_tile_image_by_gid(gid)
                sprite = pyglet.sprite.Sprite(tile, x*160, y*160, batch=self.map_batch, usage="static")
                self.map_sprites[(i, i)] = sprite

        self.player = player.Player(ad.TWIN_IMG, 0, 0)
        self.window.push_handlers(self.player)

        self.twin = twin.Twin(
            ad.EVIL_TWIN_IMG,
            0, 1000,
            evil=True
        )

        self.critters = [
            critter.Critter(ad.CRITTER_ANIMATION)
            for i in range(100)
        ]

        self.background = pyglet.sprite.Sprite(ad.BACKGROUND_IMG, -2000, -2000)

        self.particles = [
            particle.Particle(
                ad.EXPLOSION_ANIMATION, 
                -self.WINDOW_WIDTH*0.5 + self.WINDOW_WIDTH * random.random(),
                -self.WINDOW_HEIGHT*0.5 + self.WINDOW_HEIGHT * random.random(),
                self.WINDOW_WIDTH,
                self.WINDOW_HEIGHT,
                self.player
            ) 
            for i in range(0, 200)
        ]

        for i, p in enumerate(self.particles):
            p.frame_index = i % len(ad.explosion_images)

        self.game_state = GameState()

        ### HUD
        self.hud_batch = pyglet.graphics.Batch()
        self._hud_x = -self.WINDOW_WIDTH // 2 + 48
        self._hud_y = self.WINDOW_HEIGHT // 2 - 48
        self.hud = hud.Hud(
            self.hud_batch, 
            self._hud_x, 
            self._hud_y, 
            ad.HUD_CHICKEN,
            self.game_state)

        ### Menu
        self.menu_batch = pyglet.graphics.Batch()
        self.menu = menu.Menu([
            menu.MenuItem(text="Start", function=self._menu_start),
            menu.MenuItem(text="Restart", function=self._menu_restart),
            menu.MenuItem(text="Exit", function=self._menu_exit)
        ], self.menu_batch, self.WINDOW_WIDTH, self.WINDOW_HEIGHT)
        self._show_menu()

    def _restart(self):
        self.player = player.Player(ad.TWIN_IMG, 0, 0)
        self.window.push_handlers(self.player)

        self.twin = twin.Twin(
            ad.EVIL_TWIN_IMG,
            0, 1000,
            evil=True
        )

        self.critters = [
            critter.Critter(ad.CRITTER_ANIMATION)
            for i in range(100)
        ]

        self.background = pyglet.sprite.Sprite(ad.BACKGROUND_IMG, -2000, -2000)

        self.particles = [
            particle.Particle(
                ad.EXPLOSION_ANIMATION, 
                -self.WINDOW_WIDTH*0.5 + self.WINDOW_WIDTH * random.random(),
                -self.WINDOW_HEIGHT*0.5 + self.WINDOW_HEIGHT * random.random(),
                self.WINDOW_WIDTH,
                self.WINDOW_HEIGHT,
                self.player
            ) 
            for i in range(0, 200)
        ]

        for i, p in enumerate(self.particles):
            p.frame_index = i % len(ad.explosion_images)

        self.game_state = GameState()
        self.hud.update(self.game_state)

    def _menu_start(self):
        self.paused = False
        self.menu.hide()
        self.window.pop_handlers()
    
    def _menu_restart(self):
        self._menu_start()
        self.window.pop_handlers()
        self._restart()
        self.player.set_window_dimensions(self.WINDOW_WIDTH, self.WINDOW_HEIGHT)

    def _menu_exit(self):
        self.window.close()

    def _show_menu(self):
        self.menu.show()
        self.window.push_handlers(self.menu)
        self.paused = True

    def configure_window(self):
        self.window = pyglet.window.Window(
            self.WINDOW_WIDTH,
            self.WINDOW_HEIGHT,
            caption=self.GAME_TITLE,
            resizable=True,
            vsync=True
        )

        self.window.set_minimum_size(*self.WINDOW_MIN_SIZE)

        fps = pyglet.window.FPSDisplay(self.window)

        @self.window.event
        def on_resize(width, height):
            print('The window was resized to %dx%d' % (width, height))
            aspect_ratio = self.WINDOW_HEIGHT / self.WINDOW_WIDTH
            new_height = width * aspect_ratio
            glMatrixMode(gl.GL_PROJECTION)
            glLoadIdentity()
            glViewport(0, 0, width, int(new_height))
            gluOrtho2D(-0.5*self.WINDOW_WIDTH, 0.5*self.WINDOW_WIDTH, -0.5*self.WINDOW_HEIGHT, 0.5*self.WINDOW_HEIGHT)
            glMatrixMode(gl.GL_MODELVIEW)

            self.player.set_window_dimensions(width, height)

        self.window.on_resize = on_resize

        @self.window.event
        def on_key_press(symbol, modifiers):
            if symbol == key.ESCAPE:
                self._show_menu()

            return pyglet.event.EVENT_HANDLED

        @self.window.event
        def on_draw():
            glClearColor(0.5, 0.5, 0.8, 1.0)
            self.window.clear()
            glColor3f(1.0, 1.0, 1.0)

            glLoadIdentity()

            glTranslatef(-self.player.x, -self.player.y, 0.0)

            self.map_batch.draw()

            self.player.draw()

            self.twin.draw()

            for critter in self.critters:
                critter.draw()

            for particle in self.particles:
                particle.draw()

            glLoadIdentity()

            # HUD
            self.hud_batch.draw()

            if self.menu.showing:
                self.menu_batch.draw()

            fps.draw()

        @self.window.event
        def on_mouse_press(x, y, button, modifiers):
            i = 0
            x += self.player.x - self.WINDOW_WIDTH * 0.5
            y += self.player.y - self.WINDOW_HEIGHT * 0.5
            for critter in self.critters+[]:
                dist_betw = (x - critter.x) * (x - critter.x) + (y - critter.y) * (y - critter.y)
                if dist_betw < 100:
                    del self.critters[i]
                    self.game_state.critters_saved += 1
                    self.hud.update(self.game_state)
                else:
                    i += 1

            self.player.start_beam()

        @self.window.event
        def on_mouse_release(x, y, button, modifiers):
            self.player.stop_beam()


        cursor = self.window.get_system_mouse_cursor(self.window.CURSOR_CROSSHAIR)
        self.window.set_mouse_cursor(cursor)

    # --------------------------------------------------------------------------
    # Game updatessssssd
    # --------------------------------------------------------------------------
    def update(self, dt: float):

        if self.menu.showing:
            self.menu.update(dt)

        if self.paused:
            return

        for particle in self.particles:
            for twin in (self.player, self.twin):
                if twin.beam_on:
                    angle_betw = math.atan2(particle.y - twin.y, particle.x - twin.x)
                    
                    if abs(angle_betw - (twin.rotation * -math.pi / 180.0)) < 30.0 * math.pi / 180.0:
                        if (particle.x - twin.x)*(particle.x - twin.x) + (particle.y - twin.y)*(particle.y - twin.y) < twin.beam_length*twin.beam_length:
                            particle.visible = False
                            break
                    else:
                        particle.visible = True

        i = 0
        for critter in self.critters+[]:
            deleted = False
            for twin in (self.player, self.twin):
                if twin.beam_on:
                    angle_betw = math.atan2(critter.y - twin.y, critter.x - twin.x)
                    
                    if abs(angle_betw - (twin.rotation * -math.pi / 180.0)) < 30.0 * math.pi / 180.0:
                        if (critter.x - twin.x)*(critter.x - twin.x) + (critter.y - twin.y)*(critter.y - twin.y) < twin.beam_length*twin.beam_length:
                            if twin.evil:
                                twin.target_x = critter.x
                                twin.target_y = critter.y

                            critter.damage(twin.evil)

                if critter.hp[True] <= 0:
                    critter.visible = False
                    self.game_state.critters_killed += 1
                    self.hud.update(self.game_state)
                    del self.critters[i]
                    deleted = True
                    break
                elif critter.hp[False] <= 0:
                    critter.visible = False
                    self.game_state.critters_saved += 1
                    self.hud.update(self.game_state)
                    del self.critters[i]
                    deleted = True
                    break

            if not deleted:
                i += 1

        self.player.update(dt)
        self.twin.update(dt)
        for critter in self.critters:
            critter.update(dt)


# --------------------------------------------------------------------------
# Main entry point
# --------------------------------------------------------------------------
def run():
    game = Game()

    pyglet.clock.schedule_interval(game.update, 1 / 60.)

    glEnable(GL_BLEND)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    glDisable(GL_DEPTH_TEST)

    pyglet.app.run()
